import hashlib
import json
import os

from concurrent.futures import ThreadPoolExecutor
from typing import Dict
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    FAST_BAKERY_URL,
    get_pinned_conda_libs,
)
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.plugins.pypi.conda_environment import CondaEnvironment
from .fast_bakery import FastBakery, FastBakeryApiResponse, FastBakeryException
from metaflow.plugins.aws.batch.batch_decorator import BatchDecorator
from metaflow.plugins.kubernetes.kubernetes_decorator import KubernetesDecorator
from metaflow.plugins.pypi.conda_decorator import CondaStepDecorator
from metaflow.plugins.pypi.pypi_decorator import PyPIStepDecorator

BAKERY_METAFILE = ".imagebakery-cache"

import json
import os
import fcntl
from functools import wraps
from concurrent.futures import ThreadPoolExecutor


# TODO - ensure that both @conda/@pypi are not assigned to the same step


def cache_request(cache_file):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            call_args = kwargs.copy()
            call_args.update(zip(func.__code__.co_varnames, args))
            call_args.pop("self", None)
            cache_key = hashlib.md5(
                json.dumps(call_args, sort_keys=True).encode("utf-8")
            ).hexdigest()

            try:
                with open(cache_file, "r") as f:
                    cache = json.load(f)
                    if cache_key in cache:
                        return FastBakeryApiResponse(cache[cache_key])
            except (FileNotFoundError, json.JSONDecodeError):
                cache = {}

            result = func(*args, **kwargs)

            try:
                with open(cache_file, "r+") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    try:
                        f.seek(0)
                        cache = json.load(f)
                    except json.JSONDecodeError:
                        cache = {}

                    cache[cache_key] = result.response

                    f.seek(0)
                    f.truncate()
                    json.dump(cache, f)
            except FileNotFoundError:
                # path to cachefile might not exist.
                os.makedirs(os.path.dirname(cache_file), exist_ok=True)
                with open(cache_file, "w") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    json.dump({cache_key: result.response}, f)

            return result

        return wrapper

    return decorator


class DockerEnvironmentException(MetaflowException):
    headline = "Ran into an error while setting up the environment"

    def __init__(self, msg):
        super(DockerEnvironmentException, self).__init__(msg)


class DockerEnvironment(MetaflowEnvironment):
    TYPE = "fast-bakery"
    _filecache = None

    def __init__(self, flow):
        self.skipped_steps = set()
        self.flow = flow

        self.bakery = FastBakery(url=FAST_BAKERY_URL)
        self.results = {}

    def set_local_root(self, local_root):
        self.local_root = local_root

    def decospecs(self):
        return ("conda", "fast_bakery_internal") + super().decospecs()

    def validate_environment(self, echo, datastore_type):
        self.datastore_type = datastore_type
        self.echo = echo

        # Avoiding circular imports.
        from metaflow.plugins import DATASTORES

        self.datastore = [d for d in DATASTORES if d.TYPE == self.datastore_type][0]

    def init_environment(self, echo):
        self.skipped_steps = {
            step.name
            for step in self.flow
            if not any(
                isinstance(deco, (BatchDecorator, KubernetesDecorator))
                for deco in step.decorators
            )
        }

        steps_to_bake = [
            step for step in self.flow if step.name not in self.skipped_steps
        ]
        if steps_to_bake:
            echo("Baking container image(s) ...")
            self.results = self._bake(steps_to_bake, echo)
            for step in self.flow:
                for d in step.decorators:
                    if isinstance(d, (BatchDecorator, KubernetesDecorator)):
                        d.attributes["image"] = self.results[step.name].container_image
                        d.attributes["executable"] = self.results[step.name].python_path
            echo("Container image(s) baked!")

        if self.skipped_steps:
            self.delegate = CondaEnvironment(self.flow)
            self.delegate.set_local_root(self.local_root)
            self.delegate.validate_environment(echo, self.datastore_type)
            self.delegate.init_environment(echo, self.skipped_steps)

    def _bake(self, steps, echo) -> Dict[str, FastBakeryApiResponse]:
        metafile_path = get_fastbakery_metafile_path(self.local_root, self.flow.name)

        @cache_request(metafile_path)
        def _cached_bake(
            python=None, pypi_packages=None, conda_packages=None, base_image=None
        ):
            self.bakery._reset_payload()
            self.bakery.python_version(python)
            self.bakery.pypi_packages(pypi_packages)
            self.bakery.conda_packages(conda_packages)
            self.bakery.base_image(base_image)
            # self.bakery.ignore_cache()
            try:
                res = self.bakery.bake()
                if res.baking_stats:
                    echo(
                        "baked image in: %s milliseconds"
                        % res.baking_stats.solver_stats.duration_ms
                    )
                return res
            except FastBakeryException as ex:
                raise DockerEnvironmentException(str(ex))

        def prepare_step(step):
            base_image = next(
                (
                    d.attributes.get("image")
                    for d in step.decorators
                    if isinstance(d, (KubernetesDecorator))
                ),
                None,
            )
            dependencies = next(
                (
                    d
                    for d in step.decorators
                    if isinstance(d, (CondaStepDecorator, PyPIStepDecorator))
                ),
                None,
            )
            python = next(
                (
                    d.attributes["python"]
                    for d in step.decorators
                    if isinstance(d, CondaStepDecorator)
                ),
                None,
            )
            pypi_deco = next(
                (d for d in step.decorators if isinstance(d, PyPIStepDecorator)), None
            )
            # if pypi decorator is set and user has specified a python version, we must create a new environment.
            # otherwise rely on the base environment
            if pypi_deco is not None:
                python = (
                    pypi_deco.attributes["python"]
                    if pypi_deco.is_attribute_user_defined("python")
                    else None
                )

            packages = get_pinned_conda_libs(python, self.datastore_type)
            packages.update(dependencies.attributes["packages"] if dependencies else {})

            return {
                "python": python,
                "pypi_packages": (
                    packages if isinstance(dependencies, PyPIStepDecorator) else None
                ),
                "conda_packages": (
                    packages if isinstance(dependencies, CondaStepDecorator) else None
                ),
                "base_image": base_image,
            }

        with ThreadPoolExecutor() as executor:
            return {
                step.name: _cached_bake(**args)
                for step, args in zip(steps, executor.map(prepare_step, steps))
            }

    def executable(self, step_name, default=None):
        if step_name in self.skipped_steps:
            return self.delegate.executable(step_name, default)
        # default is set to the right executable
        if default is not None:
            return default
        if default is None and step_name in self.results:
            # try to read pythonpath from results. This can happen immediately after baking.
            return self.results[step_name].python_path
        # we lack a default and baking results. fallback to parent executable.
        return super().executable(step_name, default)

    def interpreter(self, step_name):
        if step_name in self.skipped_steps:
            return self.delegate.interpreter(step_name)
        return None

    def is_disabled(self, step):
        for decorator in step.decorators:
            # @conda decorator is guaranteed to exist thanks to self.decospecs
            if decorator.name in ["conda", "pypi"]:
                # handle @conda/@pypi(disabled=True)
                disabled = decorator.attributes["disabled"]
                return str(disabled).lower() == "true"
        return False

    def pylint_config(self):
        config = super().pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    def bootstrap_commands(self, step_name, datastore_type):
        if step_name in self.skipped_steps:
            return self.delegate.bootstrap_commands(step_name, datastore_type)
        # Bootstrap conda and execution environment for step
        # we set the environment flag for skipping bootstrap dependencies, as these are
        # provided in all baked images.
        return [
            "export METAFLOW_SKIP_INSTALL_DEPENDENCIES=$FASTBAKERY_IMAGE",
        ] + super().bootstrap_commands(step_name, datastore_type)


def get_fastbakery_metafile_path(local_root, flow_name):
    return os.path.join(local_root, flow_name, BAKERY_METAFILE)



from urllib.parse import urlparse

import requests
import time
import random
import os
import sys


def read_mf_config():
    # this should be overridden with the resolved remote config here:
    # obp-python-packages/ob-metaflow-extensions/metaflow_extensions/outerbounds/__init__.py
    from metaflow.metaflow_config_funcs import METAFLOW_CONFIG

    return METAFLOW_CONFIG


def get_token_url_and_headers(url_path):
    from metaflow.metaflow_config import (
        SERVICE_HEADERS,
        SERVICE_URL,
    )
    from metaflow.metaflow_config import SERVICE_HEADERS, SERVICE_URL

    # Infer auth host from metadata service URL, unless it has been
    # specified explicitly. Take the MDS host and replace first part of
    # the domain name with `auth.`. All our deployments follow this scheme
    # anyways.
    #
    auth_host = "auth." + urlparse(SERVICE_URL).hostname.split(".", 1)[1]

    authServer = read_mf_config().get("OBP_AUTH_SERVER", auth_host)
    assert url_path.startswith("/")
    url = "https://" + authServer + url_path
    headers = SERVICE_HEADERS
    return url, headers


def get_token(url_path):
    from metaflow.exception import MetaflowException

    retry_delay = 1
    max_retries = os.environ.get("METAFLOW_EXT_HTTP_MAX_RETRIES", "10")
    attempt = 0
    while attempt < int(max_retries):
        attempt = attempt + 1
        try:
            url, headers = get_token_url_and_headers(url_path)
            r = requests.get(url, headers=headers)
            r.raise_for_status()
            token_info = r.json()
            return token_info
        except requests.exceptions.HTTPError as e:
            raise MetaflowException(repr(e))
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            # ConnectionErrors are generally temporary errors like DNS resolution failures,
            # timeouts etc.
            print(
                "received error of type {}. Retrying...".format(type(e)),
                e,
                file=sys.stderr,
            )
            time.sleep(retry_delay)
            retry_delay *= 2  # Double the delay for the next attempt
            retry_delay += random.uniform(0, 1)  # Add jitter
            retry_delay = min(retry_delay, 10)

    raise MetaflowException("exceeded retries from get_token")



from metaflow.cards import Markdown, Table, VegaChart
from functools import wraps
import threading
from datetime import datetime
from metaflow import current
from metaflow.cards import Table, Markdown, VegaChart, Image
import time
from typing import List, Dict, Any, Union
import re
import os
import uuid
import json
import sys
from tempfile import TemporaryDirectory
from subprocess import check_output, Popen
from datetime import datetime, timedelta
from functools import wraps
from collections import namedtuple

# Card plot styles
MEM_COLOR = "#0c64d6"
GPU_COLOR = "#ff69b4"

NVIDIA_TS_FORMAT = "%Y/%m/%d %H:%M:%S"


DRIVER_VER = re.compile(b"Driver Version: (.+?) ")
CUDA_VER = re.compile(b"CUDA Version:(.*) ")

MONITOR_FIELDS = [
    "timestamp",
    "gpu_utilization",
    "memory_used",
    "memory_total",
]

MONITOR = """nvidia-smi --query-gpu=pci.bus_id,timestamp,utilization.gpu,memory.used,memory.total --format=csv,noheader,nounits -l {interval};"""
ProcessUUID = namedtuple("ProcessUUID", ["uuid", "start_time", "end_time"])


def _get_uuid(time_duration=600):
    frmt_str = "%Y-%m-%d-%H-%M-%S"
    # Create a datetime range between the timerange values using current date as start date and time_duration as end date
    start_date = datetime.now()
    end_date = start_date + timedelta(seconds=time_duration)
    datetime_range = start_date = (
        datetime.now().strftime(frmt_str) + "_" + end_date.strftime(frmt_str)
    )
    uuid_str = uuid.uuid4().hex.replace("-", "") + "_" + datetime_range
    return ProcessUUID(uuid_str, start_date, end_date)


class AsyncProcessManager:
    """
    This class is responsible for managing the nvidia SMI subprocesses
    """

    processes: Dict[str, Dict] = {
        # "procid": {
        #     "proc": subprocess.Popen,
        #     "started": time.time()
        # }
    }

    @classmethod
    def _register_process(cls, procid, proc):
        cls.processes[procid] = {
            "proc": proc,
            "started": time.time(),
        }

    @classmethod
    def get(cls, procid):
        proc_dict = cls.processes.get(procid, None)
        if proc_dict is not None:
            return proc_dict["proc"], proc_dict["started"]
        return None, None

    @classmethod
    def spawn(cls, procid, cmd, file):
        proc = Popen(cmd, stdout=file)
        cls._register_process(procid, proc)

    @classmethod
    def remove(cls, procid, delete_item=True):
        if procid in cls.processes:
            if cls.processes[procid]["proc"].stdout is not None:
                cls.processes[procid]["proc"].stdout.close()
            cls.processes[procid]["proc"].terminate()
            cls.processes[procid]["proc"].wait()
            if delete_item:
                del cls.processes[procid]

    @classmethod
    def cleanup(cls):
        for procid in cls.processes:
            cls.remove(procid, delete_item=False)
        cls.processes.clear()

    @classmethod
    def is_running(cls, procid):
        if procid not in cls.processes:
            return False
        return cls.processes[procid]["proc"].poll() is None


def _parse_timestamp(timestamp):
    try:
        ts = timestamp.split(".")[0]
        return datetime.strptime(ts, NVIDIA_TS_FORMAT)
    except ValueError:
        return None


class GPUMonitor:
    """
    The `GPUMonitor` class is designed to monitor GPU usage.

    When an instance of `GPUMonitor` is created, it initializes with a specified `interval` and `duration`.
    The `duration` is the timeperiod it will run the NVIDIA SMI command for and the `interval` is the timeperiod between each reading.
    The class exposes a `_monitor_update_thread` method which runs as a background thread that continuously updates the GPU usage readings.
    It will keep running unitl the `_finished` flag is set to `True`.

    The class will statefully manage the the spawned NVIDI-SMI processes.
    It will start a new NVIDI-SMI process after the current one has ran for the specified `duration`.
    At a time this class will only maintain readings for the `_current_process` and will have all the aggregated
    readings for the past processes stored in the `_past_readings` dictionary.
    When a process finishes completion, the readings are appended to the `_past_readings` dictionary and a new process is started.

    If the caller of this class wishes to read the GPU usage, they can call the `read` method which will return the readings in a dictionary format.
    The `read` method will aggregate the readings from the `_current_readings` and `_past_readings`.
    """

    _started_processes: List[ProcessUUID] = []

    _current_process: Union[ProcessUUID, None] = None

    _current_readings: Dict[str, Any] = {}

    _past_readings: Dict[str, Any] = {}

    def __init__(self, interval=1, duration=300) -> None:
        self._tempdir = TemporaryDirectory(prefix="gpu_card_monitor", dir="./")
        self._interval = interval
        self._duration = duration
        self._finished = False

    @property
    def _current_file(self):
        if self._current_process is None:
            return None
        return os.path.join(self._tempdir.name, self._current_process.uuid + ".csv")

    def get_file_name(self, uuid):
        return os.path.join(self._tempdir.name, uuid + ".csv")

    def create_new_monitor(self):
        uuid = _get_uuid(self._duration)
        file = open(self.get_file_name(uuid.uuid), "w")
        cmd = MONITOR.format(interval=self._interval, time_duration=self._duration)
        AsyncProcessManager.spawn(uuid.uuid, ["bash", "-c", cmd], file)
        self._started_processes.append(uuid)
        self._current_process = uuid
        return uuid

    def clear_current_monitor(self):
        if self._current_process is None:
            return
        AsyncProcessManager.remove(self._current_process.uuid)
        self._current_process = None

    def current_process_has_ended(self):
        if self._current_process is None:
            return True
        return datetime.now() > self._current_process.end_time

    def current_process_is_running(self):
        if self._current_process is None:
            return False
        return AsyncProcessManager.is_running(self._current_process.uuid)

    def _read_monitor(self):
        """
        Reads the monitor file and returns the readings in a dictionary format
        """
        all_readings = []
        if self._current_file is None:
            return None
        # Extract everything from the CVS File and store it in a list of dictionaries
        all_fields = ["gpu_id"] + MONITOR_FIELDS
        with open(self._current_file, "r") as _monitor_out:
            for line in _monitor_out.readlines():
                data = {}
                fields = [f.strip() for f in line.split(",")]
                if len(fields) == len(all_fields):
                    # strip subsecond resolution from timestamps that doesn't align across devices
                    for idx, _f in enumerate(all_fields):
                        data[_f] = fields[idx]
                    all_readings.append(data)
                else:
                    # expect that the last line may be truncated
                    break

        # Convert to dictionary format
        devdata = {}
        for reading in all_readings:
            gpu_id = reading["gpu_id"]
            if "timestamp" not in reading:
                continue
            if _parse_timestamp(reading["timestamp"]) is None:
                continue
            reading["timestamp"] = reading["timestamp"].split(".")[0]
            if gpu_id not in devdata:
                devdata[gpu_id] = {}

            for i, field in enumerate(MONITOR_FIELDS):
                if field not in devdata[gpu_id]:
                    devdata[gpu_id][field] = []
                devdata[gpu_id][field].append(reading[field])
        return devdata

    def _update_readings(self):
        """
        Core update function that checks if the current process has ended and if so, it will create a new monitor
        otherwise sets the current readings to the readings from the monitor file
        """
        if self.current_process_has_ended() or not self.current_process_is_running():
            self._update_past_readings()
            self.clear_current_monitor()
            self.create_new_monitor()
            # Sleep for 1 seconds to allow the new process to start and we can make a reading
            time.sleep(1)

        readings = self._read_monitor()
        if readings is None:
            return
        self._current_readings = readings

    @staticmethod
    def _make_full_reading(current, past):
        if current is None:
            return past
        for gpu_id in current:
            if gpu_id not in past:
                past[gpu_id] = {}
            for field in MONITOR_FIELDS:
                if field not in past[gpu_id]:
                    past[gpu_id][field] = []
                past[gpu_id][field].extend(current[gpu_id][field])
        return past

    def read(self):
        return self._make_full_reading(
            self._current_readings, json.loads(json.dumps(self._past_readings))
        )

    def _update_past_readings(self):
        if self._current_readings is None:
            return
        self._past_readings = self._make_full_reading(
            self._current_readings, json.loads(json.dumps(self._past_readings))
        )
        self._current_readings = None

    def cleanup(self):
        self._finished = True
        AsyncProcessManager.cleanup()
        self._tempdir.cleanup()

    def _monitor_update_thread(self):
        while not self._finished:
            self._update_readings()
            time.sleep(self._interval)


def _get_ts_range(_range):
    if _range == "":
        return "*No readings available*"
    return "*Time range of charts: %s*" % _range


def _update_utilization(results, md_dict):
    for device, data in results["profile"].items():
        if device not in md_dict:
            print(
                "Device %s not found in the GPU card layout. Skipping..." % device,
                file=sys.stderr,
            )
            continue
        md_dict[device]["gpu"].update(
            "%2.1f%%" % max(map(float, data["gpu_utilization"]))
        )
        md_dict[device]["memory"].update("%dMB" % max(map(float, data["memory_used"])))


def _update_charts(results, md_dict):
    for device, data in results["profile"].items():
        try:
            if device not in md_dict:
                continue
            gpu_plot, mem_plot, ts_range = profile_plots(
                device,
                data["timestamp"],
                data["gpu_utilization"],
                data["memory_used"],
                data["memory_total"],
            )
            md_dict[device]["gpu"].update(gpu_plot)
            md_dict[device]["memory"].update(mem_plot)
            md_dict[device]["reading_duration"].update(_get_ts_range(ts_range))
        except ValueError as e:
            # This is thrown when the date is unparsable. We can just safely ignore this.
            print("ValueError: Could not parse date \n%s" % str(e), file=sys.stderr)


# This code is adapted from: https://github.com/outerbounds/monitorbench
class GPUProfiler:
    def __init__(self, interval=1, monitor_batch_duration=200):
        self.driver_ver, self.cuda_ver, self.error = self._read_versions()
        (
            self.interconnect_data,
            self.interconnect_legend,
        ) = self._read_multi_gpu_interconnect()
        if self.error:
            self.devices = []
            return
        else:
            self.devices = self._read_devices()
            self._monitor = GPUMonitor(
                interval=interval, duration=monitor_batch_duration
            )
            self._monitor_thread = threading.Thread(
                target=self._monitor._monitor_update_thread, daemon=True
            )
            self._monitor_thread.start()
            self._interval = interval

        self._card_comps = {"max_utilization": {}, "charts": {}, "reading_duration": {}}
        self._card_created = False

    def finish(self):
        ret = {
            "error": self.error,
            "cuda_version": self.cuda_ver,
            "driver_version": self.driver_ver,
        }
        if self.error:
            return ret
        else:
            ret["devices"] = self.devices
            ret["profile"] = self._monitor.read()
            ret["interconnect"] = {
                "data": self.interconnect_data,
                "legend": self.interconnect_legend,
            }
            self._monitor.cleanup()
            return ret

    def _make_reading(self):
        ret = {
            "error": self.error,
            "cuda_version": self.cuda_ver,
            "driver_version": self.driver_ver,
        }
        if self.error:
            return ret
        else:
            ret["devices"] = self.devices
            ret["profile"] = self._monitor.read()
            ret["interconnect"] = {
                "data": self.interconnect_data,
                "legend": self.interconnect_legend,
            }
            return ret

    def _update_card(self):
        if len(self.devices) == 0:
            current.card["gpu_profile"].clear()
            current.card["gpu_profile"].append(
                Markdown("## GPU profile failed: %s" % self.error)
            )
            current.card["gpu_profile"].refresh()

            return

        while True:
            readings = self._make_reading()
            if readings is None:
                print("GPU Profiler readings are none", file=sys.stderr)
                time.sleep(self._interval)
                continue
            _update_utilization(readings, self._card_comps["max_utilization"])
            _update_charts(readings, self._card_comps["charts"])
            current.card["gpu_profile"].refresh()
            time.sleep(self._interval)

    def _setup_card(self, artifact_name):
        from metaflow import current

        results = self._make_reading()
        els = current.card["gpu_profile"]

        def _drivers():
            els.append(Markdown("## Drivers"))
            els.append(
                Table(
                    [[results["cuda_version"], results["driver_version"]]],
                    headers=["NVidia driver version", "CUDA version"],
                )
            )

        def _devices():
            els.append(Markdown("## Devices"))
            rows = [
                [d["device_id"], d["name"], d["memory"]] for d in results["devices"]
            ]
            els.append(Table(rows, headers=["Device ID", "Device type", "GPU memory"]))

        def _interconnect():
            if results["interconnect"]["data"] and results["interconnect"]["legend"]:
                els.append(Markdown("## Interconnect"))
                interconnect_data = results["interconnect"]["data"]
                rows = list(interconnect_data.values())
                rows = [list(transpose_row) for transpose_row in list(zip(*rows))]
                els.append(Table(rows, headers=list(interconnect_data.keys())))
                els.append(Markdown("#### Legend"))
                els.append(
                    Table(
                        [list(results["interconnect"]["legend"].values())],
                        headers=list(results["interconnect"]["legend"].keys()),
                    )
                )

        def _utilization():
            els.append(Markdown("## Maximum utilization"))
            rows = {}
            for d in results["devices"]:
                rows[d["device_id"]] = {
                    "gpu": Markdown("0%"),
                    "memory": Markdown("0MB"),
                }
            _rows = [[Markdown(k)] + list(v.values()) for k, v in rows.items()]
            els.append(
                Table(data=_rows, headers=["Device ID", "Max GPU %", "Max memory"])
            )
            els.append(
                Markdown(f"Detailed data saved in an artifact `{artifact_name}`")
            )
            return rows

        def _plots():
            els.append(Markdown("## GPU utilization and memory usage over time"))

            rows = {}
            for d in results["devices"]:
                gpu_plot, mem_plot, ts_range = profile_plots(
                    d["device_id"], [], [], [], []
                )
                rows[d["device_id"]] = {
                    "gpu": VegaChart(gpu_plot),
                    "memory": VegaChart(mem_plot),
                    "reading_duration": Markdown(_get_ts_range(ts_range)),
                }
            for k, v in rows.items():
                els.append(Markdown("### GPU Utilization for device : %s" % k))
                els.append(v["reading_duration"])
                els.append(
                    Table(
                        data=[
                            [Markdown("GPU Utilization"), v["gpu"]],
                            [Markdown("Memory usage"), v["memory"]],
                        ]
                    )
                )
            return rows

        _drivers()
        _devices()
        _interconnect()
        self._card_comps["max_utilization"] = _utilization()
        self._card_comps["charts"] = _plots()

    def _read_versions(self):
        def parse(r, s):
            return r.search(s).group(1).strip().decode("utf-8")

        try:
            out = check_output(["nvidia-smi"])
            return parse(DRIVER_VER, out), parse(CUDA_VER, out), None
        except FileNotFoundError:
            return None, None, "nvidia-smi not found"
        except AttributeError:
            return None, None, "nvidia-smi output is unexpected"
        except:
            return None, None, "nvidia-smi error"

    def _read_devices(self):
        out = check_output(
            [
                "nvidia-smi",
                "--query-gpu=name,pci.bus_id,memory.total",
                "--format=csv,noheader",
            ]
        )
        return [
            dict(
                zip(("name", "device_id", "memory"), (x.strip() for x in l.split(",")))
            )
            for l in out.decode("utf-8").splitlines()
        ]

    def _read_multi_gpu_interconnect(self):
        """
        parse output of `nvidia-smi tomo -m`, such as this sample:

            GPU0    GPU1    CPU Affinity    NUMA Affinity
            GPU0     X      NV2     0-23            N/A
            GPU1    NV2      X      0-23            N/A

        returns two dictionaries describing multi-GPU topology:
            data: {index: [GPU0, GPU1, ...], GPU0: [X, NV2, ...], GPU1: [NV2, X, ...], ...}
            legend_items: {X: 'Same PCI', NV2: 'NVLink 2', ...}
        """
        try:
            import re

            ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")

            out = check_output(["nvidia-smi", "topo", "-m"])
            rows = out.decode("utf-8").split("\n")

            header = ansi_escape.sub("", rows[0]).split("\t")[1:]
            data = {}
            data["index"] = []
            data |= {k: [] for k in header}

            for i, row in enumerate(rows[1:]):
                row = ansi_escape.sub("", row).split()
                if len(row) == 0:
                    continue
                if row[0].startswith("GPU"):
                    data["index"].append(row[0])
                    for key, val in zip(header, row[1:]):
                        data[key].append(val)
                elif row[0].startswith("Legend"):
                    break

            legend_items = {}
            for legend_row in rows[i:]:
                if legend_row == "" or legend_row.startswith("Legend"):
                    continue
                res = legend_row.strip().split(" = ")
                legend_items[res[0].strip()] = res[1].strip()

            return data, legend_items

        except:
            return None, None


class gpu_profile:
    def __init__(
        self,
        include_artifacts=True,
        artifact_prefix="gpu_profile_",
        interval=1,
    ):
        self.include_artifacts = include_artifacts
        self.artifact_prefix = artifact_prefix
        self.interval = interval

    def __call__(self, f):
        @wraps(f)
        def func(s):
            prof = GPUProfiler(interval=self.interval)
            if self.include_artifacts:
                setattr(s, self.artifact_prefix + "num_gpus", len(prof.devices))

            current.card["gpu_profile"].append(
                Markdown("# GPU profile for `%s`" % current.pathspec)
            )
            current.card["gpu_profile"].append(
                Markdown(
                    "_Started at: %s_"
                    % datetime.now().astimezone().strftime("%Y-%m-%dT%H:%M:%S %z")
                )
            )
            prof._setup_card(self.artifact_prefix + "data")
            current.card["gpu_profile"].refresh()
            update_thread = threading.Thread(target=prof._update_card, daemon=True)
            update_thread.start()

            try:
                f(s)
            finally:
                try:
                    results = prof.finish()
                except:
                    results = {"error": "couldn't read profiler results"}
                if self.include_artifacts:
                    setattr(s, self.artifact_prefix + "data", results)

        from metaflow import card

        return card(type="blank", id="gpu_profile", refresh_interval=self.interval)(
            func
        )


def translate_to_vegalite(
    tstamps,
    vals,
    description,
    y_label,
    legend,
    line_color=None,
    percentage_format=False,
):
    # Preprocessing for Vega-Lite
    # Assuming tstamps is a list of datetime objects and vals is a list of values
    data = [{"tstamps": str(t), "vals": v} for t, v in zip(tstamps, vals)]

    # Base Vega-Lite spec
    vega_lite_spec = {
        "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
        "description": description,
        "data": {"values": data},
        "width": 600,
        "height": 400,
        "encoding": {
            "x": {"field": "tstamps", "type": "temporal", "axis": {"title": "Time"}},
            "y": {
                "field": "vals",
                "type": "quantitative",
                "axis": {
                    "title": y_label,
                    **({"format": "%"} if percentage_format else {}),
                },
            },
        },
        "layer": [
            {
                "mark": {
                    "type": "line",
                    "color": line_color if line_color else "blue",
                    "tooltip": True,
                    "description": legend,  # Adding legend as description
                },
                "encoding": {"tooltip": [{"field": "tstamps"}, {"field": "vals"}]},
            }
        ],
    }

    return vega_lite_spec


def profile_plots(device_id, ts, gpu, mem_used, mem_total):
    tstamps = [datetime.strptime(t, NVIDIA_TS_FORMAT) for t in ts]
    gpu = [i / 100 for i in list(map(float, gpu))]
    mem = [float(used) / float(total) for used, total in zip(mem_used, mem_total)]
    time_stamp_range = ""
    if len(tstamps) > 1:
        max_time = max(tstamps).strftime(NVIDIA_TS_FORMAT)
        min_time = min(tstamps).strftime(NVIDIA_TS_FORMAT)
        time_stamp_range = "%s to %s" % (min_time, max_time)

    gpu_plot = translate_to_vegalite(
        tstamps,
        gpu,
        "GPU utilization",
        "GPU utilization",
        "device: %s" % device_id,
        line_color=GPU_COLOR,
        percentage_format=True,
    )
    mem_plot = translate_to_vegalite(
        tstamps,
        mem,
        "Percentage Memory utilization",
        "Percentage Memory utilization",
        "device: %s" % device_id,
        line_color=MEM_COLOR,
        percentage_format=True,
    )
    return gpu_plot, mem_plot, time_stamp_range


if __name__ == "__main__":
    prof = GPUProfiler(monitor_batch_duration=10)

    def _write_json_file(data, filename):
        with open(filename, "w") as f:
            json.dump(data, f, indent=4)

    import time

    for i in range(15):
        time.sleep(1)
        _write_json_file(prof._monitor.read(), "gpu_profile.json")

    print(json.dumps(prof.finish()))



import os
import shlex
import atexit
import json
import math
import time

from metaflow import util

from metaflow.metaflow_config import (
    SERVICE_INTERNAL_URL,
    SERVICE_HEADERS,
    DEFAULT_METADATA,
    DATASTORE_SYSROOT_S3,
    DATATOOLS_S3ROOT,
    KUBERNETES_SANDBOX_INIT_SCRIPT,
    OTEL_ENDPOINT,
    DEFAULT_SECRETS_BACKEND_TYPE,
    AWS_SECRETS_MANAGER_DEFAULT_REGION,
    S3_SERVER_SIDE_ENCRYPTION,
    S3_ENDPOINT_URL,
)
from metaflow.metaflow_config_funcs import config_values

from metaflow.mflog import (
    export_mflog_env_vars,
    bash_capture_logs,
    BASH_SAVE_LOGS,
    get_log_tailer,
    tail_logs,
)

from .snowpark_client import SnowparkClient
from .snowpark_exceptions import SnowparkException
from .snowpark_job import SnowparkJob

# Redirect structured logs to $PWD/.logs/
LOGS_DIR = "$PWD/.logs"
STDOUT_FILE = "mflog_stdout"
STDERR_FILE = "mflog_stderr"
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)


class Snowpark(object):
    def __init__(
        self,
        datastore,
        metadata,
        environment,
        client_credentials,
    ):
        self.datastore = datastore
        self.metadata = metadata
        self.environment = environment
        self.snowpark_client = SnowparkClient(**client_credentials)
        atexit.register(lambda: self.job.kill() if hasattr(self, "job") else None)

    def _job_name(self, user, flow_name, run_id, step_name, task_id, retry_count):
        return "{user}-{flow_name}-{run_id}-{step_name}-{task_id}-{retry_count}".format(
            user=user,
            flow_name=flow_name,
            run_id=str(run_id) if run_id is not None else "",
            step_name=step_name,
            task_id=str(task_id) if task_id is not None else "",
            retry_count=str(retry_count) if retry_count is not None else "",
        )

    def _command(self, environment, code_package_url, step_name, step_cmds, task_spec):
        mflog_expr = export_mflog_env_vars(
            datastore_type=self.datastore.TYPE,
            stdout_path=STDOUT_PATH,
            stderr_path=STDERR_PATH,
            **task_spec
        )
        init_cmds = environment.get_package_commands(
            code_package_url, self.datastore.TYPE
        )
        init_expr = " && ".join(init_cmds)
        step_expr = bash_capture_logs(
            " && ".join(
                environment.bootstrap_commands(step_name, self.datastore.TYPE)
                + step_cmds
            )
        )

        # construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        # the `true` command is to make sure that the generated command
        # plays well with docker containers which have entrypoint set as
        # eval $@
        cmd_str = "true && mkdir -p %s && %s && %s && %s; " % (
            LOGS_DIR,
            mflog_expr,
            init_expr,
            step_expr,
        )
        # after the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario.
        cmd_str += "c=$?; %s; exit $c" % BASH_SAVE_LOGS
        # For supporting sandboxes, ensure that a custom script is executed before
        # anything else is executed. The script is passed in as an env var.
        cmd_str = (
            '${METAFLOW_INIT_SCRIPT:+eval \\"${METAFLOW_INIT_SCRIPT}\\"} && %s'
            % cmd_str
        )
        return shlex.split('bash -c "%s"' % cmd_str)

    def create_job(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        image=None,
        stage=None,
        compute_pool=None,
        volume_mounts=None,
        external_integration=None,
        cpu=None,
        gpu=None,
        memory=None,
        run_time_limit=None,
        env=None,
        attrs=None,
    ) -> SnowparkJob:
        if env is None:
            env = {}
        if attrs is None:
            attrs = {}

        job_name = self._job_name(
            attrs.get("metaflow.user"),
            attrs.get("metaflow.flow_name"),
            attrs.get("metaflow.run_id"),
            attrs.get("metaflow.step_name"),
            attrs.get("metaflow.task_id"),
            attrs.get("metaflow.retry_count"),
        )

        snowpark_job = (
            SnowparkJob(
                client=self.snowpark_client,
                name=job_name,
                command=self._command(
                    self.environment, code_package_url, step_name, [step_cli], task_spec
                ),
                step_name=step_name,
                step_cli=step_cli,
                task_spec=task_spec,
                code_package_sha=code_package_sha,
                code_package_url=code_package_url,
                code_package_ds=code_package_ds,
                image=image,
                stage=stage,
                compute_pool=compute_pool,
                volume_mounts=volume_mounts,
                external_integration=external_integration,
                cpu=cpu,
                gpu=gpu,
                memory=memory,
                run_time_limit=run_time_limit,
                env=env,
                attrs=attrs,
            )
            .environment_variable("METAFLOW_CODE_SHA", code_package_sha)
            .environment_variable("METAFLOW_CODE_URL", code_package_url)
            .environment_variable("METAFLOW_CODE_DS", code_package_ds)
            .environment_variable("METAFLOW_USER", attrs["metaflow.user"])
            .environment_variable("METAFLOW_SERVICE_URL", SERVICE_INTERNAL_URL)
            .environment_variable(
                "METAFLOW_SERVICE_HEADERS", json.dumps(SERVICE_HEADERS)
            )
            .environment_variable("METAFLOW_DATASTORE_SYSROOT_S3", DATASTORE_SYSROOT_S3)
            .environment_variable("METAFLOW_DATATOOLS_S3ROOT", DATATOOLS_S3ROOT)
            .environment_variable("METAFLOW_DEFAULT_DATASTORE", self.datastore.TYPE)
            .environment_variable("METAFLOW_DEFAULT_METADATA", DEFAULT_METADATA)
            .environment_variable("METAFLOW_SNOWPARK_WORKLOAD", 1)
            .environment_variable("METAFLOW_RUNTIME_ENVIRONMENT", "snowpark")
            .environment_variable(
                "METAFLOW_INIT_SCRIPT", KUBERNETES_SANDBOX_INIT_SCRIPT
            )
            .environment_variable("METAFLOW_OTEL_ENDPOINT", OTEL_ENDPOINT)
            .environment_variable(
                "SNOWFLAKE_WAREHOUSE",
                self.snowpark_client.connection_parameters.get("warehouse"),
            )
        )

        for k, v in config_values():
            if k.startswith("METAFLOW_CONDA_") or k.startswith("METAFLOW_DEBUG_"):
                snowpark_job.environment_variable(k, v)

        if DEFAULT_SECRETS_BACKEND_TYPE is not None:
            snowpark_job.environment_variable(
                "METAFLOW_DEFAULT_SECRETS_BACKEND_TYPE", DEFAULT_SECRETS_BACKEND_TYPE
            )
        if AWS_SECRETS_MANAGER_DEFAULT_REGION is not None:
            snowpark_job.environment_variable(
                "METAFLOW_AWS_SECRETS_MANAGER_DEFAULT_REGION",
                AWS_SECRETS_MANAGER_DEFAULT_REGION,
            )
        if S3_SERVER_SIDE_ENCRYPTION is not None:
            snowpark_job.environment_variable(
                "METAFLOW_S3_SERVER_SIDE_ENCRYPTION", S3_SERVER_SIDE_ENCRYPTION
            )
        if S3_ENDPOINT_URL is not None:
            snowpark_job.environment_variable(
                "METAFLOW_S3_ENDPOINT_URL", S3_ENDPOINT_URL
            )

        for name, value in env.items():
            snowpark_job.environment_variable(name, value)

        return snowpark_job

    def launch_job(self, **kwargs):
        self.job = self.create_job(**kwargs).create().execute()

    def wait(self, stdout_location, stderr_location, echo=None):
        def update_delay(secs_since_start):
            # this sigmoid function reaches
            # - 0.1 after 11 minutes
            # - 0.5 after 15 minutes
            # - 1.0 after 23 minutes
            # in other words, the user will see very frequent updates
            # during the first 10 minutes
            sigmoid = 1.0 / (1.0 + math.exp(-0.01 * secs_since_start + 9.0))
            return 0.5 + sigmoid * 30.0

        def wait_for_launch(job):
            status = job.status

            echo(
                "Task is starting (%s)..." % status,
                "stderr",
                job_id=job.id,
            )
            t = time.time()
            start_time = time.time()
            while job.is_waiting:
                new_status = job.status
                if status != new_status or (time.time() - t) > 30:
                    status = new_status
                    echo(
                        "Task is starting (%s)..." % status,
                        "stderr",
                        job_id=job.id,
                    )
                    t = time.time()
                time.sleep(update_delay(time.time() - start_time))

        _make_prefix = lambda: b"[%s] " % util.to_bytes(self.job.id)

        stdout_tail = get_log_tailer(stdout_location, self.datastore.TYPE)
        stderr_tail = get_log_tailer(stderr_location, self.datastore.TYPE)

        # 1) Loop until the job has started
        wait_for_launch(self.job)

        # 2) Tail logs until the job has finished
        tail_logs(
            prefix=_make_prefix(),
            stdout_tail=stdout_tail,
            stderr_tail=stderr_tail,
            echo=echo,
            has_log_updates=lambda: self.job.is_running,
        )

        if self.job.has_failed:
            msg = next(
                msg
                for msg in [
                    self.job.message,
                    "Task crashed.",
                ]
                if msg is not None
            )
            raise SnowparkException(
                "%s " "This could be a transient error. " "Use @retry to retry." % msg
            )
        else:
            if self.job.is_running:
                # Kill the job if it is still running by throwing an exception.
                raise SnowparkException("Task failed!")
            echo(
                "Task finished with message '%s'." % self.job.message,
                "stderr",
                job_id=self.job.id,
            )



import os
import sys
import time

from metaflow.exception import MetaflowException


CLIENT_REFRESH_INTERVAL_SECONDS = 300


class KubernetesClientException(MetaflowException):
    headline = "Kubernetes client error"


class KubernetesClient(object):
    def __init__(self):
        try:
            # Kubernetes is a soft dependency.
            from kubernetes import client, config
        except (NameError, ImportError):
            raise KubernetesClientException(
                "Could not import module 'kubernetes'.\n\nInstall Kubernetes "
                "Python package (https://pypi.org/project/kubernetes/) first.\n"
                "You can install the module by executing - "
                "%s -m pip install kubernetes\n"
                "or equivalent through your favorite Python package manager."
                % sys.executable
            )
        self._refresh_client()

    def _refresh_client(self):
        from metaflow_extensions.outerbounds.plugins.auth_server import get_token
        from kubernetes import client

        config = client.Configuration()
        token_info = get_token("/generate/k8s")
        config.host = token_info["endpoint"]
        config.api_key["authorization"] = "Bearer " + token_info["token"]
        config.verify_ssl = False  # TODO: FIX THIS
        client.Configuration.set_default(config)
        self._client = client
        self._client_refresh_timestamp = time.time()

    def get(self):
        if (
            time.time() - self._client_refresh_timestamp
            > CLIENT_REFRESH_INTERVAL_SECONDS
        ):
            self._refresh_client()

        return self._client

    def job(self, **kwargs):
        from metaflow.plugins.kubernetes.kubernetes_job import KubernetesJob

        return KubernetesJob(self, **kwargs)

    def jobset(self, **kwargs):
        from metaflow.plugins.kubernetes.kubernetes_job import KubernetesJobSet

        return KubernetesJobSet(self, **kwargs)
